comment|
Demonstration of the generation and use of an index into a lookup table.

Sample build file:

@echo off
\masm32\bin\ml /c /coff index.asm
if errorlevel 1 goto end
\masm32\bin\Link /SUBSYSTEM:CONSOLE index.obj
if errorlevel 1 goto end
del index.obj
dir index.exe
:end
|

.586
.model flat, stdcall
option casemap :none

include \masm32\include\windows.inc
include \masm32\include\kernel32.inc
includelib \masm32\lib\kernel32.lib

capal macro
local lbl
    cmp al,"a"
    jb short lbl
    cmp al,"z"
    ja short lbl
    and al,0DFh
lbl:
endm

filemax equ 1000000h   ;16 megs

ReadWholeFileBySpec     PROTO :DWORD,:DWORD
NormalizeEOLs           PROTO :DWORD
TabulateStringAddresses PROTO :DWORD,:DWORD,:BYTE
IndexedLookup           PROTO :DWORD,:DWORD,:DWORD,:DWORD,:NEAR
line_cmp                PROTO :DWORD,:DWORD
line_cmp_nocase         PROTO :DWORD,:DWORD
tree                    PROTO :DWORD,:DWORD

.data
hStdOut dd ?
AmtWritten dd ?
index_memhandle dd ?
parm dd ?
datasize dd ?
addresses_memhandle dd ?
initial_record_number dd ? ;an index into both tables: string addies and integer pairs
input_file_contents dd ?
input_line_count dd ?

testfilespec db "countries.txt",0
Canada$ db "Canada",10
Sveden$ db "Sveden",10
LiBeria$ db "LiBeria",10

report_line db "Record number 0x"
report_var dd ?
    db 13,10
report_line_ equ 22

.code
start:
    invoke GetStdHandle,STD_OUTPUT_HANDLE
    mov  hStdOut,eax
    invoke ReadWholeFileBySpec,addr testfilespec,filemax
    jz   help1
    mov  input_file_contents,eax
    invoke NormalizeEOLs,eax
    test edx,0FFFF0000h
    jnz  help2
    mov  input_line_count,edx
    shl  edx,2
    mov  datasize,edx
    invoke LocalAlloc,0,edx
    test eax,eax
    jz   help3
    mov  index_memhandle,eax
    mov  ecx,datasize
    add  ecx,eax
    invoke tree,eax,ecx  ;fill in the index table with signed short integers
    sub  eax,index_memhandle
    shr  eax,2
    mov  initial_record_number,eax
    invoke TabulateStringAddresses,input_file_contents,input_line_count,10
    jz   quit
    mov  addresses_memhandle,eax
    invoke IndexedLookup,addr Canada$,
                         eax,
                         index_memhandle,
                         initial_record_number,
                         addr line_cmp
    call report
    invoke IndexedLookup,addr Sveden$,
                         addresses_memhandle,
                         index_memhandle,
                         initial_record_number,
                         addr line_cmp
    call report
    invoke IndexedLookup,addr LiBeria$,
                         addresses_memhandle,
                         index_memhandle,
                         initial_record_number,
                         addr line_cmp_nocase
    call report

help1:    ;stub error handlers
help2:
help3:
quit:
    invoke ExitProcess,0

tree proc bottom:DWORD,top:DWORD
;Re-entrant routine, fills in a table of dwords with pairs of
;short signed integers determined by the size of the table.
;Returns eax = midpoint, which will be the entry point for the lookup routine.
local midd:DWORD
    mov  edx,bottom
    mov  eax,top
    xor  ecx,ecx
    sub  eax,edx
    shr  eax,3
    shl  eax,2         ;sets or clears the zero flag
    lea  eax,[eax+edx]
    mov  midd,eax
    mov  [eax],ecx
    jz   short @F
      invoke tree,edx,eax
    mov  edx,midd
    sub  eax,edx
    mov  [edx],ax
@@: mov  edx,midd
    add  edx,4
    cmp  edx,top
    je   short @F
      invoke tree,edx,top
    mov  edx,midd
    sub  eax,edx
    mov  [edx+2],ax
@@: mov  eax,midd
    ret
tree endp

ReadWholeFileBySpec proc filespec:DWORD,maxsize:DWORD
      ;returns eax=memory handle and edx=file size without terminator.
local hFile:DWORD , sizz:DWORD , memhandle:DWORD
    xor  eax,eax
    mov  memhandle,eax
    invoke CreateFileA,filespec,GENERIC_READ,
           FILE_SHARE_READ or FILE_SHARE_WRITE,
           eax,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,eax
    test eax,eax
    jz   short RWFBS_ret
    mov  hFile, eax
    invoke GetFileSize,eax,0
    mov  sizz, eax
    inc  eax
    jz   short RWFBS_closeret
    inc  eax            ;allow room for an appended line feed and zero
    cmp  eax,maxsize
    ja   short RWFBS_closeret
    invoke LocalAlloc,GMEM_FIXED,eax
    mov  memhandle,eax
    test eax,eax
    jnz  short RWFBS_doread
RWFBS_closeret:
    invoke CloseHandle,hFile
    jmp  short RWFBS_ret
RWFBS_doread:
    push eax         ; create a 4-byte structure for ReadFile's amount read
    mov  edx,esp
    invoke ReadFile,hFile,memhandle,sizz,edx,0
    invoke CloseHandle,hFile
    pop  edx         ;amount read
    mov  eax,memhandle
    mov  byte ptr[eax+edx],0
RWFBS_ret:
    mov  eax,memhandle
    test eax,eax
    ret
ReadWholeFileBySpec endp

TabulateStringAddresses proc uses ebx edi esi szs:DWORD,linecount:DWORD,terminator:BYTE
;If enough memory, returns eax=address of new allocation in the heap,
;which contains the addresses of the strings found in "szs".
;Returns eax=0 if not enough heap.
    xor  esi,esi
    mov  ebx,linecount
    lea  eax,[4*ebx]
    invoke LocalAlloc,GMEM_FIXED,eax
    test eax,eax
    jz   short TSA_ret
    xchg eax,edx
    mov  edi,szs
    or   ebx,-1
    mov  al,terminator
@@: inc  ebx
    cmp  ebx,linecount
    je   short @F
    mov  [edx+4*ebx],edi
    or   ecx,-1
    repne scasb
    jmp  short @B
@@: xchg eax,edx
TSA_ret:
    test eax,eax
    ret
TabulateStringAddresses endp

NormalizeEOLs proc memhandle:DWORD
      ;Input on the stack is the address of some 0-terminated text.
      ;Works <i>in situ</i>, replaces any line breaks with 0Ah.
      ;Omits empty lines, and trims every line at both ends.
      ;Returns eax=byte count and edx=line count of the output.
      ;Appends an EOL character 0Ah at the end if there wasn't one.
    push ebx
    push esi
    push edi
    xor  ebx,ebx    ;ebx will maintain the count of lines
    xor  eax,eax
    mov  esi,memhandle
    mov  edi,esi
    mov  ecx,esi    ;ecx = location just beyond the last nonspace in the output.
    mov  dl,10      ;line feed. dl will remember the last output character.
NormalizeEOLs_loop:
    lodsb
    test al,al
    jz   short NormalizeEOLs_finish
    cmp  al,13
    je   short NormalizeEOLs_cr
    cmp  al,10
    je   short NormalizeEOLs_lf
    cmp  al," "
    je   short NormalizeEOLs_sp
    cmp  al,9
    je   short NormalizeEOLs_sp
    mov  dl,al
    stosb
    mov  ecx,edi
    jmp  short NormalizeEOLs_loop
NormalizeEOLs_sp:
    cmp  dl,10
    je   short NormalizeEOLs_loop
    mov  dl,al
    stosb
    jmp  short NormalizeEOLs_loop
NormalizeEOLs_cr:
    mov  al,10
NormalizeEOLs_lf:
    cmp  dl,10
    je   short NormalizeEOLs_loop  ;don't store consecutive EOL's
    mov  edi,ecx
    stosb
    inc  ebx
    mov  dl,al
    mov  ecx,edi
    jmp  NormalizeEOLs_loop
NormalizeEOLs_finish:
    mov  al,10
    cmp  al,dl
    je   short @F
    stosb
@@: mov  byte ptr[edi],0
    xchg eax,edi
    sub  eax,memhandle
    mov  edx,ebx     ;line count
    pop  ebx
    pop  edi
    pop  esi
    ret
NormalizeEOLs endp

IndexedLookup proc item:DWORD,
                   item_locations:DWORD,
                   index_table:DWORD,
                   entry_recnum:DWORD,
                   comparer:NEAR
;If the item is found, returns eax = the 1-based record number of that
;item. If not found, returns eax=0.
    mov  edi,item_locations
    mov  ebx,entry_recnum
    shl  ebx,2
    mov  esi,index_table
IL_loop:
    push [ebx+edi]
    push item
    call comparer
    jz   short IL_success
    jc   short IL_look_higher
;IL_look_lower:
    movsx eax,word ptr[esi+ebx+2]
    jmp  short IL_adjust
IL_look_higher:
    movsx eax,word ptr[esi+ebx]
IL_adjust:
    test eax,eax
    jz   short IL_done
    add  ebx,eax
    jmp  short IL_loop
IL_success:
    xchg eax,ebx
    shr  eax,2
    inc  eax
IL_done:
    ret
IndexedLookup endp

line_cmp proc uses esi edi str1:DWORD,str2:DWORD
;case-sensitive, and returns info in the flags only.
    mov  esi,str1
    mov  edi,str2
@@: lodsb
    mov  ah,[edi]
    inc  edi
    cmp  ax,0A0Ah
    je   short @F
    cmp  al,ah
    jz   short @B
@@: ret
line_cmp endp

line_cmp_nocase proc uses esi edi str1:DWORD,str2:DWORD
;case-sensitive, and returns info in the flags only.
    mov  esi,str1
    mov  edi,str2
@@: lodsb
    mov  ah,[edi]
    inc  edi
    cmp  ax,0A0Ah
    je   short @F
    capal
    xchg al,ah
    capal
    xchg al,ah
    cmp  al,ah
    jz   short @B
@@: ret
line_cmp_nocase endp

AX2hexEAX:
    mov  dl,ah
    call AL2hexAX
    ror  eax,16
    mov  al,dl
AL2hexAX:
    mov  ah,al
    shr  ah,4
    call Nibble2hexAL
    xchg al,ah
Nibble2hexAL:
    and  al,0Fh
    cmp  al,0Ah
    sbb  al,69h
    das
    ret

report:
    call AX2hexEAX
    mov report_var,eax
    invoke WriteFile,hStdOut,addr report_line,report_line_,addr AmtWritten,0
    ret

end start
